# EnforceVars by David Bowland
# ./addons/eventscripts/enforcevars/enforcevars.py


# >>> To configure this addon please see enforcevars.cfg  (created when the addon is first loaded) <<<

"""
Enforces the values of client cheat variables and kicks or bans players with unacceptable values.
Requires only ES 2.0.0.250i+
"""

import cfglib
import cmdlib
import es
import gamethread
import langlib
import os.path
import playerlib
import random
import services
import time


import psyco
psyco.full()

###

info = es.AddonInfo()
info.name      = 'EnforceVars'
info.version   = '1'
info.basename  = 'enforcevars'
info.url       = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author    = 'SuperDave'

###

SECONDS_BETWEEN_QUERIES = 0.8 # Number of seconds between each variable query

###

config = cfglib.AddonCFG(os.path.join(es.getAddonPath(info.basename), info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')
config.text('To configure language settings for %s please see %s_languages.ini' % (info.name, info.basename))
config.text('\n')

config.text('***** Query delay option *****')
cvar_querydelay = config.cvar('enforcevars_querydelay', 15, 'Minimum number of seconds between variable queries')
config.text('\n')

config.text('***** Punishment option *****')
config.text('Players with the "enforcevars_immunity" permission will be immune from variable enforcement.')
config.text('By default said permission is registered to the root permission level.')
config.text('More information on authorization services: http://www.eventscripts.com/pages/Authorization_FAQ')
cvar_ban_time = config.cvar('enforcevars_ban_time', -1, 'Number of minutes to ban clients with conflicting variable values, 0 = permanent, -1 = kick only')
config.text('\n')

config.text('***** Enforced variable options *****')
config.text('')
config.text('Use the following server command to specify client variables to be enforced and acceptable values:')
config.text('')
config.text('enforcevars_add <variable> [value 1] [value 2] [value n]')
config.text('')
config.text('Due to the number of variables enforced querying a client variable may take several minutes.')
config.text('\n')

config.text('Place your enforcevars_add commands below:')
config.command('enforcevars_add')
config.text('\n')

config.text('Uncomment the following default options if you wish to use them:')
config.text('')
config.text('enforcevars_add ai_debug_los 0')
config.text('enforcevars_add ai_debug_shoot_positions 0')
config.text('enforcevars_add ai_drawbattlelines 0')
config.text('enforcevars_add bot_crouch 0')
config.text('enforcevars_add bot_debug 0')
config.text('enforcevars_add bot_debug_target 0')
config.text('enforcevars_add bot_dont_shoot 0')
config.text('enforcevars_add bot_freeze 0')
config.text('enforcevars_add bot_loadout 0')
config.text('enforcevars_add bot_mimic 0')
config.text('enforcevars_add bot_mimic_yaw_offset 180')
config.text('enforcevars_add bot_show_battlefront 0')
config.text('enforcevars_add bot_show_nav 0')
config.text('enforcevars_add bot_show_occupy_time 0')
config.text('enforcevars_add bot_stop 0')
config.text('enforcevars_add bot_traceview 0')
config.text('enforcevars_add bot_zombie 0')
config.text('enforcevars_add cam_command 0')
config.text('enforcevars_add cl_backspeed 400')
config.text('enforcevars_add cl_bob 0.002')
config.text('enforcevars_add cl_bobcycle 0.8')
config.text('enforcevars_add cl_bobup 0.5')
config.text('enforcevars_add cl_clock_correction 1')
config.text('enforcevars_add cl_clock_correction_adjustment_max_amount 200')
config.text('enforcevars_add cl_clock_correction_adjustment_max_offset 90')
config.text('enforcevars_add cl_clock_correction_adjustment_min_offset 10')
config.text('enforcevars_add cl_clock_correction_force_server_tick 999')
config.text('enforcevars_add cl_clock_showdebuginfo 0')
config.text('enforcevars_add cl_clockdrift_max_ms 150')
config.text('enforcevars_add cl_detailfade 400')
config.text('enforcevars_add cl_drawhud 1')
config.text('enforcevars_add cl_drawleaf -1')
config.text('enforcevars_add cl_drawmaterial ""')
config.text('enforcevars_add cl_drawshadowtexture 0')
config.text('enforcevars_add cl_entityreport 0')
config.text('enforcevars_add cl_extrapolate 1')
config.text('enforcevars_add cl_extrapolate_amount 0.25')
config.text('enforcevars_add cl_flushentitypacket 0')
config.text('enforcevars_add cl_forwardspeed 400')
config.text('enforcevars_add cl_ignorepackets 0')
config.text('enforcevars_add cl_leveloverview 0')
config.text('enforcevars_add cl_leveloverviewmarker 0')
config.text('enforcevars_add cl_maxrenderable_dist 3000')
config.text('enforcevars_add cl_overdraw_test 0')
config.text('enforcevars_add cl_particleeffect_aabb_buffer 2')
config.text('enforcevars_add cl_particles_show_bbox 0')
config.text('enforcevars_add cl_pclass ""')
config.text('enforcevars_add cl_pdump -1')
config.text('enforcevars_add cl_phys_props_enable 1')
config.text('enforcevars_add cl_phys_timescale 1.0')
config.text('enforcevars_add cl_pitchdown 89')
config.text('enforcevars_add cl_pitchup 89')
config.text('enforcevars_add cl_predict 1.0')
config.text('enforcevars_add cl_predictionlist 0')
config.text('enforcevars_add cl_showanimstate -1')
config.text('enforcevars_add cl_showanimstate_log 0')
config.text('enforcevars_add cl_showerror 0')
config.text('enforcevars_add cl_showevents 0')
config.text('enforcevars_add cl_sidespeed 400')
config.text('enforcevars_add cl_sporeclipdistance 512')
config.text('enforcevars_add cl_sun_decay_rate 0.05')
config.text('enforcevars_add cl_upspeed 320')
config.text('enforcevars_add cl_winddir 0')
config.text('enforcevars_add cl_windspeed 0')
config.text('enforcevars_add cl_wpn_sway_scale 1.0')
config.text('enforcevars_add cs_ShowStateTransitions -2')
config.text('enforcevars_add default_fov 90')
config.text('enforcevars_add demo_recordcommands 1')
config.text('enforcevars_add developer 0')
config.text('enforcevars_add disp_numiterations 1')
config.text('enforcevars_add dsp_dist_max 1440')
config.text('enforcevars_add dsp_dist_max 1440')
config.text('enforcevars_add dsp_dist_min 0')
config.text('enforcevars_add dsp_off 0')
config.text('enforcevars_add default_fov 90')
config.text('enforcevars_add demo_recordcommands 1')
config.text('enforcevars_add developer 0')
config.text('enforcevars_add disp_numiterations 1')
config.text('enforcevars_add dsp_dist_max 1440')
config.text('enforcevars_add dsp_dist_min 0')
config.text('enforcevars_add dsp_off 0')
config.text('enforcevars_add ent_messages_draw 0')
config.text('enforcevars_add fish_debug 0')
config.text('enforcevars_add fish_dormant 0')
config.text('enforcevars_add fog_color "-1 -1 -1"')
config.text('enforcevars_add fog_colorskybox "-1 -1 -1"')
config.text('enforcevars_add fog_enable 1')
config.text('enforcevars_add fog_enable_water_fog 1')
config.text('enforcevars_add fog_enableskybox 1')
config.text('enforcevars_add fog_end -1')
config.text('enforcevars_add fog_endskybox -1')
config.text('enforcevars_add fog_override 0')
config.text('enforcevars_add fog_start -1')
config.text('enforcevars_add fog_startskybox -1')
config.text('enforcevars_add g_debug_angularsensor 0')
config.text('enforcevars_add g_debug_ragdoll_removal 0')
config.text('enforcevars_add g_debug_ragdoll_visualize 0')
config.text('enforcevars_add g_debug_trackpather 0')
config.text('enforcevars_add g_debug_vehiclebase 0')
config.text('enforcevars_add g_debug_vehicledriver 0')
config.text('enforcevars_add g_debug_vehicleexit 0')
config.text('enforcevars_add g_debug_vehiclesound 0')
config.text('enforcevars_add g_jeepexitspeed 100')
config.text('enforcevars_add hidehud 0')
config.text('enforcevars_add host_sleep 0')
config.text('enforcevars_add hostage_debug 0')
config.text('enforcevars_add lightcache_maxmiss 2')
config.text('enforcevars_add mat_accelerate_adjust_exposure_down 3')
config.text('enforcevars_add mat_bloomamount_rate "0.05f"')
config.text('enforcevars_add mat_bumpbasis 0')
config.text('enforcevars_add mat_camerarendertargetoverlaysize 128')
config.text('enforcevars_add mat_debug_autoexposure 0')
config.text('enforcevars_add mat_debug_bloom 0')
config.text('enforcevars_add mat_debug_process_halfscreen 0')
config.text('enforcevars_add mat_debugalttab 0')
config.text('enforcevars_add mat_debugdepth 0')
config.text('enforcevars_add mat_depthbias_decal -262144')
config.text('enforcevars_add mat_depthbias_normal 0')
config.text('enforcevars_add mat_diffuse 1')
config.text('enforcevars_add mat_drawflat 0')
config.text('enforcevars_add mat_drawwater 1')
config.text('enforcevars_add mat_dynamic_tonemapping 1')
config.text('enforcevars_add mat_exposure_center_region_x 0.75')
config.text('enforcevars_add mat_exposure_center_region_x_flashlight 0.33')
config.text('enforcevars_add mat_exposure_center_region_y 0.80')
config.text('enforcevars_add mat_exposure_center_region_y_flashlight 0.33')
config.text('enforcevars_add mat_fillrate 0')
config.text('enforcevars_add mat_filtertextures 1')
config.text('enforcevars_add mat_force_bloom 0')
config.text('enforcevars_add mat_force_tonemap_scale 0')
config.text('enforcevars_add mat_forcedynamic 0')
config.text('enforcevars_add mat_frame_sync_enable 1')
config.text('enforcevars_add mat_frame_sync_force_texture 0')
config.text('enforcevars_add mat_hdr_manual_tonemap_rate 1')
config.text('enforcevars_add mat_hdr_tonemapscale 2.500000')
config.text('enforcevars_add mat_hdroverbrightrange 16')
config.text('enforcevars_add mat_hsv 0')
config.text('enforcevars_add mat_leafvis 0')
config.text('enforcevars_add mat_loadtextures 1')
config.text('enforcevars_add mat_luxels 0')
config.text('enforcevars_add mat_measurefillrate 0')
config.text('enforcevars_add mat_mipmaptextures 1')
config.text('enforcevars_add mat_norendering 0')
config.text('enforcevars_add mat_normalmaps 0')
config.text('enforcevars_add mat_normals 0')
config.text('enforcevars_add mat_proxy 0')
config.text('enforcevars_add mat_reversedepth 0')
config.text('enforcevars_add mat_show_texture_memory_usage 0')
config.text('enforcevars_add mat_showcamerarendertarget 0')
config.text('enforcevars_add mat_showframebuffertexture 0')
config.text('enforcevars_add mat_showlightmappage -1')
config.text('enforcevars_add mat_showlowresimage 0')
config.text('enforcevars_add mat_showmiplevels 0')
config.text('enforcevars_add mat_showwatertextures 0')
config.text('enforcevars_add mat_slopescaledepthbias_decal 0')
config.text('enforcevars_add mat_slopescaledepthbias_normal 0')
config.text('enforcevars_add mat_softwarelighting 0')
config.text('enforcevars_add mat_softwareskin 0')
config.text('enforcevars_add mat_stub 0')
config.text('enforcevars_add mat_surfaceid 0')
config.text('enforcevars_add mat_surfacemat 0')
config.text('enforcevars_add mat_texture_limit -1')
config.text('enforcevars_add mat_viewportscale 1.0')
config.text('enforcevars_add mat_wireframe 0')
config.text('enforcevars_add mat_yuv 0')
config.text('enforcevars_add mem_force_flush 0')
config.text('enforcevars_add mod_forcedata 1')
config.text('enforcevars_add mod_load_anims_async 0')
config.text('enforcevars_add mod_load_mesh_async 0')
config.text('enforcevars_add mod_load_vcollide_async 0')
config.text('enforcevars_add mod_lock_mdls_on_load 0')
config.text('enforcevars_add mod_test_mesh_not_available 0')
config.text('enforcevars_add mod_test_not_available 0')
config.text('enforcevars_add mod_test_verts_not_available 0')
config.text('enforcevars_add mod_trace_load 0')
config.text('enforcevars_add nav_edit 0')
config.text('enforcevars_add nav_quicksave 0')
config.text('enforcevars_add nav_show_approach_points 0')
config.text('enforcevars_add nav_show_danger 0')
config.text('enforcevars_add nav_show_player_counts 0')
config.text('enforcevars_add net_blockmsg 0')
config.text('enforcevars_add net_droppackets 0')
config.text('enforcevars_add net_fakelag 0')
config.text('enforcevars_add net_fakeloss 0')
config.text('enforcevars_add next 0')
config.text('enforcevars_add npc_ally_deathmessage 1')
config.text('enforcevars_add particle_simulateoverflow 0')
config.text('enforcevars_add pwatchent -1')
config.text('enforcevars_add pwatchvar ""')
config.text('enforcevars_add r_ambientfraction 0')
config.text('enforcevars_add r_ambientlightingonly 0')
config.text('enforcevars_add r_aspectratio 0')
config.text('enforcevars_add r_avglight 1')
config.text('enforcevars_add r_avglightmap 0')
config.text('enforcevars_add r_ClipAreaPortals 1')
config.text('enforcevars_add r_colorstaticprops 0')
config.text('enforcevars_add r_debugcheapwater 0')
config.text('enforcevars_add r_debugrandomstaticlighting 0')
config.text('enforcevars_add r_DispBuildable 0')
config.text('enforcevars_add r_DispWalkable 0')
config.text('enforcevars_add r_DrawBeams 1')
config.text('enforcevars_add r_drawbrushmodels 1')
config.text('enforcevars_add r_drawclipbrushes 0')
config.text('enforcevars_add r_drawdecals 1')
config.text('enforcevars_add r_drawdetailprops 1')
config.text('enforcevars_add r_DrawDisp 1')
config.text('enforcevars_add r_drawentities 1')
config.text('enforcevars_add r_drawfullskybox 1')
config.text('enforcevars_add r_drawleaf -1')
config.text('enforcevars_add r_drawlightcache 0')
config.text('enforcevars_add r_drawlightinfo 0')
config.text('enforcevars_add r_drawlights 0')
config.text('enforcevars_add r_DrawModelLightOrigin 0')
config.text('enforcevars_add r_drawmodelstatsoverlay 0')
config.text('enforcevars_add r_drawmodelstatsoverlaydistance 500')
config.text('enforcevars_add r_drawopaquerenderables 1')
config.text('enforcevars_add r_drawopaqueworld 1')
config.text('enforcevars_add r_drawothermodels 1')
config.text('enforcevars_add r_drawparticles 1')
config.text('enforcevars_add r_DrawPortals 0')
config.text('enforcevars_add r_DrawRain 1')
config.text('enforcevars_add r_drawrenderboxes 0')
config.text('enforcevars_add r_drawropes 1')
config.text('enforcevars_add r_drawskybox 1')
config.text('enforcevars_add r_drawsprites 1')
config.text('enforcevars_add r_drawstaticprops 1')
config.text('enforcevars_add r_drawtranslucentrenderables 1')
config.text('enforcevars_add r_drawtranslucentworld 1')
config.text('enforcevars_add r_drawvgui 1')
config.text('enforcevars_add r_drawviewmodel 1')
config.text('enforcevars_add r_drawworld 1')
config.text('enforcevars_add r_dscale_basefov 90')
config.text('enforcevars_add r_dscale_fardist 2000')
config.text('enforcevars_add r_dscale_farscale 4')
config.text('enforcevars_add r_dscale_neardist 100')
config.text('enforcevars_add r_dscale_nearscale 1')
config.text('enforcevars_add r_entity -1')
config.text('enforcevars_add r_eyewaterepsilon "7.0f"')
config.text('enforcevars_add r_farz -1')
config.text('enforcevars_add r_flashlightconstant 0.0')
config.text('enforcevars_add r_flashlightfar 750.0')
config.text('enforcevars_add r_flashlightfov 45.0')
config.text('enforcevars_add r_flashlightlinear 100.0')
config.text('enforcevars_add r_flashlightlockposition 0')
config.text('enforcevars_add r_flashlightnear 1.0')
config.text('enforcevars_add r_flashlightoffsetx 10.0')
config.text('enforcevars_add r_flashlightoffsety -20.0')
config.text('enforcevars_add r_flashlightoffsetz 24.0')
config.text('enforcevars_add r_flashlightquadratic 0.0')
config.text('enforcevars_add r_flashlightvisualizetrace 0')
config.text('enforcevars_add r_itemblinkmax 0')
config.text('enforcevars_add r_itemblinkrate 4')
config.text('enforcevars_add r_lightcache_numambientsamples 162')
config.text('enforcevars_add r_lightcachecenter 1')
config.text('enforcevars_add r_lightinterp 5')
config.text('enforcevars_add r_lightmap -1')
config.text('enforcevars_add r_lightstyle -1')
config.text('enforcevars_add r_lockpvs 0')
config.text('enforcevars_add r_mapextents 16384')
config.text('enforcevars_add r_modellodscale 1')
config.text('enforcevars_add r_modelwireframedecal 0')
config.text('enforcevars_add r_newflashlight 1')
config.text('enforcevars_add r_nohw 0')
config.text('enforcevars_add r_nosw 0')
config.text('enforcevars_add r_novis 0')
config.text('enforcevars_add r_occlusionspew 0')
config.text('enforcevars_add r_partition_level -1')
config.text('enforcevars_add r_portalsopenall 0')
config.text('enforcevars_add r_PortalTestEnts 1')
config.text('enforcevars_add r_propsmaxdist 1200')
config.text('enforcevars_add r_radiosity 4')
config.text('enforcevars_add r_rainalpha 0.4')
config.text('enforcevars_add r_rainalphapow 0.8')
config.text('enforcevars_add r_raindensity 0.001')
config.text('enforcevars_add r_RainHack 0')
config.text('enforcevars_add r_rainlength "0.1f"')
config.text('enforcevars_add r_RainProfile 0')
config.text('enforcevars_add r_RainRadius 1500')
config.text('enforcevars_add r_RainSideVel 130')
config.text('enforcevars_add r_RainSimulate 1')
config.text('enforcevars_add r_rainspeed "600.0f"')
config.text('enforcevars_add r_RainSplashPercentage 20')
config.text('enforcevars_add r_rainwidth 0.5')
config.text('enforcevars_add r_screenfademaxsize 0')
config.text('enforcevars_add r_screenfademinsize 0')
config.text('enforcevars_add r_shadowids 0')
config.text('enforcevars_add r_shadows_gamecontrol 1')
config.text('enforcevars_add r_shadowwireframe 0')
config.text('enforcevars_add r_showenvcubemap 0')
config.text('enforcevars_add r_skin 0')
config.text('enforcevars_add r_skybox 1')
config.text('enforcevars_add r_SnowColorBlue 200')
config.text('enforcevars_add r_SnowColorGreen 175')
config.text('enforcevars_add r_SnowColorRed 150')
config.text('enforcevars_add r_SnowDebugBox 0')
config.text('enforcevars_add r_SnowEnable 1')
config.text('enforcevars_add r_SnowEndAlpha 255')
config.text('enforcevars_add r_SnowEndSize 0')
config.text('enforcevars_add r_SnowFallSpeed 1.5')
config.text('enforcevars_add r_SnowInsideRadius 256')
config.text('enforcevars_add r_SnowOutsideRadius 1024')
config.text('enforcevars_add r_SnowParticles 500')
config.text('enforcevars_add r_SnowPosScale 1')
config.text('enforcevars_add r_SnowRayEnable 1')
config.text('enforcevars_add r_SnowRayLength "8192.0f"')
config.text('enforcevars_add r_SnowRayRadius 256')
config.text('enforcevars_add r_SnowSpeedScale 1')
config.text('enforcevars_add r_SnowStartAlpha 25')
config.text('enforcevars_add r_SnowStartSize 1')
config.text('enforcevars_add r_SnowWindScale 0.0035')
config.text('enforcevars_add r_SnowZoomOffset "384.0f"')
config.text('enforcevars_add r_SnowZoomRadius "512.0f"')
config.text('enforcevars_add r_TransitionSensitivity 6')
config.text('enforcevars_add r_visocclusion 0')
config.text('enforcevars_add r_visualizelighttraces 0')
config.text('enforcevars_add r_visualizelighttracesshowfulltrace 0')
config.text('enforcevars_add r_visualizetraces 0')
config.text('enforcevars_add showbudget_texture 0')
config.text('enforcevars_add showtriggers 0')
config.text('enforcevars_add singlestep 0')
config.text('enforcevars_add snd_foliage_db_loss 4')
config.text('enforcevars_add snd_gain 1')
config.text('enforcevars_add snd_gain_max 1')
config.text('enforcevars_add snd_gain_min 0')
config.text('enforcevars_add snd_obscured_gain_dB -2')
config.text('enforcevars_add snd_refdb 60')
config.text('enforcevars_add snd_refdist 36')
config.text('enforcevars_add snd_show 0')
config.text('enforcevars_add snd_showstart 0')
config.text('enforcevars_add snd_visualize 0')
config.text('enforcevars_add soundscape_fadetime 3.0')
config.text('enforcevars_add sv_cheats 0')
config.text('enforcevars_add sv_netvisdist 10000')
config.text('enforcevars_add sv_noclipduringpause 0')
config.text('enforcevars_add sv_pushaway_hostage_force 20000')
config.text('enforcevars_add sv_pushaway_max_hostage_force 1000')
config.text('enforcevars_add sv_pushaway_max_player_force 10000')
config.text('enforcevars_add sv_pushaway_player_force 200000')
config.text('enforcevars_add sv_showanimstate -1')
config.text('enforcevars_add sv_showanimstate_log 0')
config.text('enforcevars_add sv_showhitboxes -1')
config.text('enforcevars_add sv_showlagcompensation 0')
config.text('enforcevars_add vcollide_wireframe 0')
config.text('enforcevars_add vgui_drawtree 0')
config.text('enforcevars_add viewmodel_fov 54')
config.text('')
config.text('The following client variable can be triggered by some cutstom maps.')
config.text('If you use custom maps you may need to omit this variables from the enforced variable list')
config.text('enforcevars_add mat_fullbright 0')
config.text('\n')

config.text('Use the following command to remove variables from the list of variables to be enforced:')
config.text('')
config.text('enforcevars_remove <variable>')
config.text('')
config.text('Due to the fact enforced variables can be specified above, this command will not be used by most users.')


config.write()

###

ini = cfglib.AddonINI(os.path.join(es.getAddonPath(info.basename), info.basename + '_languages.ini'))

ini.setInitialComments(['# ./addons/eventscripts/%(basename)s/%(basename)s_languages.ini' % {'basename': info.basename},
 '# >>> For general addon configuration see ' + info.basename + '.cfg <<<', '', '# DO NOT translate words with $ or # in front!',
 '# For a list of language abbreviations please see ./addons/eventscripts/_libs/python/deflangs.ini'])

ini.addGroup('kick variable')
ini.addValueToGroup('kick variable', 'en', 'Client variable "$variable" value "$client" conflicting with server value "$server"')

ini.write()
lang_text = langlib.Strings(ini)

###

class ClientQuery(object):
   """ Class for handling the enforcement of client variables """
   auth = None

   def __init__(self):
      # Setup our data structures for enforced variables and player query times
      self.enforcedvars = {}
      self.players = {}

      # Players authorized at the ROOT level are immune from variable enforcement
      if services.isRegistered('auth'):
         self.auth = services.use('auth')
         self.auth.registerCapability('enforcevars_immunity', self.auth.ROOT)

      # Register server commands for designating variables to be enforced
      cmdlib.registerServerCommand('enforcevars_add', self.addVar, 'enforcevars_add <variable> [value]\nAdds a variable to the list EnforceVars watches')
      cmdlib.registerServerCommand('enforcevars_remove', self.addVar, 'enforcevars_remove <variable>\nRemoves a variable to the list EnforceVars watches')

   @staticmethod
   def unregisterCommands():
      """ Unregisters the EnforceVars server command """
      cmdlib.unregisterServerCommand('enforcevars_add')
      cmdlib.unregisterServerCommand('enforcevars_remove')

   """ Player-related functions """

   def removePlayer(self, userid):
      """ Removes a userid from the database of queried players """
      userid = int(userid)
      if userid in self.players:
         del self.players[userid]

   def trimPlayers(self):
      """ Trims players out of the database who are outside the query delay """
      now = time.time()
      for userid in filter(lambda userid: now - self.players[userid] >= cvar_querydelay, self.players):
         del self.players[userid]

   """ Variable enforcement list functions """

   def addVar(self, args):
      """
      enforcevars_add <variable> [value 1] [value 2] [value n]
      """
      if len(args) > 0:
         accepted_values = []
         # We want to compare each client variable with its server equivalent if possible
         if es.exists('variable', args[0]):
            accepted_values += [es.ServerVar(args[0])]

         # Add the provided values to the acceptable values list
         if len(args) > 1:
            accepted_values += args[1:]

         # If no values were provided and no server variable found
         # we have nothing to compare the client's value with
         if not accepted_values:
            raise NameError, "Variable '%s' does not exist on server" % args[0]
         # Store the acceptable values for the current variable
         self.enforcedvars[args[0]] = accepted_values

      else:
         es.dbgmsg(0, 'Syntax: enforcevars_add <variable> [value 1] [value 2] [value n]')

   def removeVar(self, args):
      """
      enforcevars_remove <variable>
      """
      if len(args) == 1:
         if args[0] in self.enforcedvars:
            del self.enforcedvars[args[0]]

      else:
         es.dbgmsg(0, 'Syntax: enforcevars_remove <variable>')

   """ Query functions  """

   def triggerQuery(self, userid):
      """ Starts the query of enforced variables """
      userid = int(userid)
      # If the player is immune from variable enforcement we don't query the variables
      if self.auth and self.auth.isUseridAuthorized(userid, 'enforcevars_immunity'):
         return

      # We only query variables after a certain amount of time has passed
      now = time.time()
      if now - self.players.get(userid, 0) < cvar_querydelay:
         return
      self.players[userid] = now

      # Send a query for each variable
      # Using a random order makes query times less predictable
      for x, var in enumerate(random.sample(list(self.enforcedvars), len(self.enforcedvars))):
         gamethread.delayedname(x * SECONDS_BETWEEN_QUERIES, 'enforcevars', self.sendQuery, (userid, var))

   def resultCallback(self, userid, var, clientvalue):
      """ Compares the queried value with acceptable values """
      # If the variables is not enforced we aren't concerned with its value
      if var not in self.enforcedvars:
         return

      # Compare the client's value to the acceptable server values
      userid = int(userid)
      servervalues = self.enforcedvars[var]
      if clientvalue in servervalues:
         # The values match but we don't want to start another query until this one has completely finished
         self.players[userid] = time.time()
         return

      # Gather language information and call the "punish" function
      lang = playerlib.getPlayer(userid).lang
      displayvalues = '" or "'.join(map(str, servervalues))
      tokens = {'variable': var, 'client': clientvalue, 'server': displayvalues}
      self.punish(userid, lang_text('kick variable', tokens, lang))

   def fail(self, userid, var, status):
      """ Log query failure """
      # If the variables is not one we are enforcing we don't care about the query failure
      if var in self.enforcedvars:
         es.dbgmsg(0, 'EnforceVars: Failure querying variable "%s" (%s) from %s (%s)' %
          (var, status, es.getplayername(userid), es.getplayersteamid(userid)))

   @staticmethod
   def punish(userid, text):
      """ Kicks and bans a player with the provided text """
      if cvar_ban_time >= 0:
         es.server.queuecmd('banid %s %s;writeid' % (cvar_ban_time, userid))
      es.server.queuecmd('kickid %s %s' % (userid, text))

   @staticmethod
   def sendQuery(userid, var):
      """ Sends a query for a single client variable """
      if es.getuserid(userid):
         es.queryclientvar(userid, var)


query = ClientQuery()

###

def load():
   config.execute()


def es_map_start(event_var):
   query.trimPlayers()


def es_player_variable(event_var):
   # If for some strange reason the userid is not valid (connected) we ignore this line
   if not es.getuserid(event_var['userid']):
      return

   # Variable query callback
   if event_var['status'] == 'success':
      query.resultCallback(event_var['userid'], event_var['variable'], event_var['value'])

   # Query failed
   else:
      query.fail(event_var['userid'], event_var['variable'], event_var['status'])


def player_spawn(event_var):
   if event_var['es_steamid'] not in ('STEAM_ID_PENDING', 'BOT'):
      # We query client variables on each spawn
      query.triggerQuery(event_var['userid'])


def player_death(event_var):
   # We also query client varaibles every time the client gets a kill to make sure nothing has changed
   if event_var['attacker'] not in (event_var['userid'], '0') and event_var['es_attackersteamid'] != 'BOT':
      query.triggerQuery(event_var['attacker'])


def player_disconnect(event_var):
   query.removePlayer(event_var['userid'])


def unload():
   query.unregisterCommands()
   gamethread.cancelDelayed('enforcevars')

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)